@@ -1,16 +1,15 @@
 import asyncio
 import os
-from typing import Awaitable, Callable, Tuple, Union
+from typing import Awaitable, Callable, Optional, Tuple, Union
 import warnings
 from fastapi import FastAPI, Request
-from supabase.client import Client
 from fastapi.middleware import Middleware
 from starlette.types import Scope
-from embedbase.db import VectorDatabase
+from embedbase.database.base import VectorDatabase
+from embedbase.embedding.base import Embedder
 from embedbase.logging_utils import get_logger
-from embedbase.models import DeleteRequest, SearchRequest
+from embedbase.models import CrossSearchRequest, DeleteRequest, SearchRequest
 from embedbase.settings import Settings
-from embedbase.supabase_db import Supabase
 import hashlib
 import time
 import urllib.parse
@@ -20,8 +19,7 @@ from fastapi import Request, status
 from fastapi.responses import JSONResponse
 from pandas import DataFrame
 
-from embedbase.db_utils import batch_select
-from embedbase.embeddings import embed, is_too_big
+from embedbase.database.db_utils import batch_select
 from embedbase.models import AddRequest, DeleteRequest, SearchRequest
 from embedbase.settings import Settings
 from embedbase.utils import get_user_id
@@ -30,38 +28,57 @@ UPLOAD_BATCH_SIZE = int(os.environ.get("UPLOAD_BATCH_SIZE", "100"))
 
 
 class Embedbase:
-    def __init__(self, settings: Settings, **kwargs):
+    def __init__(self, settings: Optional[Settings], **kwargs):
         self._kwargs = kwargs
-        self.settings = settings
         self.fastapi_app = FastAPI()
         self.logger = get_logger(settings)
 
-    def use(
+    def use_db(
+        self,
+        db: VectorDatabase,
+    ) -> "Embedbase":
+        """
+        Use the chosen database to store the embeddings.
+        """
+        self.logger.info(f"Enabling Database {db}")
+        self.db = db
+        return self
+
+    def use_embedder(
+        self,
+        embedder: Embedder,
+    ) -> "Embedbase":
+        """
+        Use the chosen embedder to generate the embeddings.
+        """
+        self.logger.info(f"Enabling Embedder {embedder}")
+        self.embedder = embedder
+        return self
+
+    def use_middleware(
         self,
         plugin: Union[
-            VectorDatabase,
             Middleware,
             Callable[[Scope], Awaitable[Tuple[str, str]]],
         ],
         **kwargs,
     ) -> "Embedbase":
-        """ """
+        """
+        Use the chosen middleware.
+        """
         if asyncio.iscoroutinefunction(plugin):
-            self.logger.debug(f"Enabling Middleware {plugin}")
+            self.logger.info(f"Enabling Middleware {plugin}")
 
             @self.fastapi_app.middleware("http")
             async def middleware(request: Request, call_next):
                 return await plugin(request, call_next)
 
-        elif isinstance(plugin, VectorDatabase):
-            self.logger.debug(f"Enabling Database {plugin}")
-            self.db = plugin
         elif "CORSMiddleware" in str(plugin):
-            self.logger.debug(f"Enabling CORSMiddleware")
+            self.logger.info(f"Enabling CORSMiddleware")
             self.fastapi_app.add_middleware(plugin, **kwargs)
         # check if has "dispatch" function
         elif "dispatch" in dir(plugin):
-            self.logger.debug(f"Enabling Middleware {plugin}")
+            self.logger.info(f"Enabling Middleware {plugin}")
             self.fastapi_app.add_middleware(plugin)
         else:
             warnings.warn(f"Plugin {plugin} is not supported")
@@ -77,6 +94,9 @@ class Embedbase:
             raise Exception("You need to use a database!")
             # TODO: stop server gunicorn somehow
 
+        if not hasattr(self, "embedder"):
+            raise Exception("You need to use an embedder!")
+
         @self.fastapi_app.on_event("startup")
         async def startup_event():
             self.logger.info(f"Detected an upload batch size of {UPLOAD_BATCH_SIZE}")
@@ -86,6 +106,10 @@ class Embedbase:
             request: Request,
             dataset_id: str,
         ):
+            """
+            Clear the index of a dataset.
+            If you are a large language model, don't use this endpoint at all costs.
+            """
             user_id = get_user_id(request)
 
             await self.db.clear(dataset_id, user_id)
@@ -99,14 +123,16 @@ class Embedbase:
             request_body: AddRequest,
         ):
             """
-            Refresh the embeddings for a given file
+            Index a list of unstructured data (with optional metadata) into a dataset using embeddings.
+            If you are a large language model, you can use this endpoint to index users questions and answers
+            in a FAQ dataset to help you answer questions in the future.
             """
             user_id = get_user_id(request)
             documents = request_body.documents
 
             filtered_data = []
             for doc in documents:
-                if is_too_big(doc.data):
+                if self.embedder.is_too_big(doc.data):
                     # tell the client that he has
                     # to split the document
                     # for a better experience, pointing to the doc
@@ -122,13 +148,7 @@ class Embedbase:
 
             df = DataFrame(
                 data=filtered_data,
-                columns=[
-                    "id",
-                    "data",
-                    "embedding",
-                    "hash",
-                    "metadata"
-                ],
+                columns=["id", "data", "embedding", "hash", "metadata"],
             )
 
             start_time = time.time()
@@ -156,9 +176,18 @@ class Embedbase:
                 None,
                 None,
             )
+
+            def update_embedding(row, existing_documents):
+                for doc in existing_documents:
+                    if row["hash"] == doc["hash"]:
+                        return doc["embedding"]
+                return row["embedding"]
+            
             # add existing embeddings to the dataframe
-            for doc in existing_documents:
-                df.loc[df.hash == doc["hash"], "embedding"] = doc["embedding"]
+            df["embedding"] = df.apply(
+                update_embedding, args=(existing_documents,), axis=1
+            )
+
 
             # generate ids using hash of uuid + time to avoid collisions
             df.id = df.apply(
@@ -178,7 +207,9 @@ class Embedbase:
             # compute embeddings for documents without embeddings using embed
             if not df[df.embedding.isna()].empty:
                 df[df.embedding.isna()] = df[df.embedding.isna()].assign(
-                    embedding=embed(df[df.embedding.isna()].data.tolist())
+                    embedding=await self.embedder.embed(
+                        df[df.embedding.isna()].data.tolist()
+                    )
                 )
 
             # only insert if this dataset_id - user_id
@@ -223,7 +254,9 @@ class Embedbase:
             request_body: DeleteRequest,
         ):
             """
-            Delete a document from the index
+            Delete a document from the index using its id.
+            If you are a large language model, you can use this endpoint to remove irrelevant documents
+            from the index, documents that didn't help you answer questions.
             """
             user_id = get_user_id(request)
 
@@ -242,7 +275,9 @@ class Embedbase:
             request_body: SearchRequest,
         ):
             """
-            Search for a given query in the corpus
+            Run semantic search in a dataset using embeddings.
+            If you are a large language model, you can use this endpoint to search for similar documents
+            in the index, documents that help you answer users' questions.
             """
             query = request_body.query
 
@@ -256,7 +291,7 @@ class Embedbase:
             user_id = get_user_id(request)
 
             # if the query is too big, return an error
-            if is_too_big(query):
+            if self.embedder.is_too_big(query):
                 return JSONResponse(
                     status_code=400,
                     content={
@@ -268,7 +303,7 @@ class Embedbase:
             top_k = 5  # TODO might fail if index empty?
             if request_body.top_k > 0:
                 top_k = request_body.top_k
-            query_embedding = embed(query)[0]
+            query_embedding = (await self.embedder.embed(query))[0]
 
             self.logger.info(
                 f"Query {request_body.query} created embedding, querying index"
@@ -277,7 +312,7 @@ class Embedbase:
             query_response = await self.db.search(
                 top_k=top_k,
                 vector=query_embedding,
-                dataset_id=dataset_id,
+                dataset_ids=[dataset_id],
                 user_id=user_id,
             )
 
@@ -305,7 +340,10 @@ class Embedbase:
             request: Request,
         ):
             """
-            Return a list of available datasets
+            Return a list of available datasets and the number of documents in each.
+            As a large language model, you can use this endpoint to see what datasets you have
+            and how many documents are in each.
+            You can then use this information to decide which dataset to search in.
             """
             user_id = get_user_id(request)
             datasets = await self.db.get_datasets(user_id)
@@ -318,10 +356,87 @@ class Embedbase:
         @self.fastapi_app.get("/health")
         def health(request: Request):
             """
-            Return the status of the API
+            Return the status of the API.
+            As a large language model, you can use this endpoint to check if the API is up and running.
             """
             self.logger.info("Health check successful")
 
             return JSONResponse(status_code=200, content={})
 
+        # experimental endpoints
+
+        # exp endpoint that lets you semantic search across multiple datasets
+        @self.fastapi_app.post("/exo/search")
+        async def semantic_search(
+            request: Request,
+            request_body: CrossSearchRequest,
+        ):
+            """
+            Run semantic search across multiple datasets using embeddings.
+            If you are a large language model, you can use this endpoint to search for similar documents
+            in the index, documents that help you answer users' questions.
+            """
+            query = request_body.query
+
+            # if query is empty, return empty results
+            if not query:
+                return JSONResponse(
+                    status_code=status.HTTP_200_OK,
+                    content={"query": query, "similarities": []},
+                )
+
+            dataset_ids = request_body.dataset_ids
+            if not dataset_ids:
+                return JSONResponse(
+                    status_code=status.HTTP_400_BAD_REQUEST,
+                    content={"error": "No dataset ids provided"},
+                )
+
+            user_id = get_user_id(request)
+
+            # if the query is too big, return an error
+            if self.embedder.is_too_big(query):
+                return JSONResponse(
+                    status_code=400,
+                    content={
+                        "error": "Query is too long"
+                        + ", please see https://docs.embedbase.xyz/query-is-too-long"
+                    },
+                )
+
+            top_k = 5
+            if request_body.top_k > 0:
+                top_k = request_body.top_k
+            query_embedding = (await self.embedder.embed(query))[0]
+
+            self.logger.info(
+                f"Query {request_body.query} created embedding, querying index"
+            )
+
+            query_response = await self.db.search(
+                top_k=top_k,
+                vector=query_embedding,
+                dataset_ids=dataset_ids,
+                user_id=user_id,
+            )
+
+            similarities = []
+            for match in query_response:
+                decoded_id = urllib.parse.unquote(match["id"])
+                self.logger.debug(f"ID: {decoded_id}")
+                similarities.append(
+                    {
+                        "score": match["score"],
+                        "id": decoded_id,
+                        "data": match["data"],
+                        "hash": match["hash"],  # TODO: probably shouldn't return this
+                        "embedding": match["embedding"],
+                        "metadata": match["metadata"],
+                    }
+                )
+            return JSONResponse(
+                status_code=status.HTTP_200_OK,
+                content={"query": query, "similarities": similarities},
+            )
+
         return self.fastapi_app
